#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
# Barcerlona Supercomputing Center - Centro Nacional de Supercomputacion
#        Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)
#             -
#             -
#
import sys, os, argparse, subprocess
from dqc_wrapper.test00_file_format import FileFormat
from dqc_wrapper.test01_standard_compliance import StandardComplianceTest
from dqc_wrapper.test02_spatial_completeness import SpatialCompletenessTest
from dqc_wrapper.test03_temporal_completeness import TemporalCompleteness
from dqc_wrapper.test04_spatial_consistency import SpatialConsistencyTest
from dqc_wrapper.test05_temporal_consistency import TemporalConsistencyTest
from dqc_wrapper.test06_valid_ranges import ValidRangesTest


test2class = {
    "file_format": FileFormat,
    "standard_compliance": StandardComplianceTest,
    "spatial_completeness": SpatialCompletenessTest,
    "temporal_completness": TemporalCompleteness,
    "spatial_consistency": SpatialConsistencyTest,
    "temporal_consistency": TemporalConsistencyTest,
    "valid_ranges": ValidRangesTest
}


def create_parser(args):
    help_file = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "checker_help.txt")
    with open(help_file, 'r') as file:
        help = file.read()
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--test", help="Type of test",
                        choices=test2class.keys(), required=True)
    parser.add_argument("--input", help="File or folder", required=True)
    parser.add_argument("--type", help="Type of file")
    parser.add_argument("--temporal_range", help="Time interval (dates)")
    parser.add_argument("--space_interval", help="Spatial grid")
    parser.add_argument("--time_interval", help="Time interval")
    parser.add_argument("--var", help="Variable")
    return parser.parse_args(args)


class Checker:

    def __init__(self, name, args):
        self.version = "0.0.1"
        self.name = name
        self.args = args

    def __version__(self):
        print("\nData Quality Checker Wrapper Version: " + self.version + "\n")

    def main(self):
        self.__version__()
        if self.args.test == "spatial_consistency":
            with open(os.devnull, 'w') as fp:
                subprocess.Popen([
                    'utils/check_spatial_consistency.bash %s %s %s' % (
                    self.args.type, self.args.space_interval,
                    self.args.input)], shell=True, stderr=fp)
        elif self.args.test == "valid_ranges":
            test = ValidRangesTest(self.args)
            test.check(k_sd_REF=10)
        else:
            test = test2class[self.args.test](self.args)
            test.main()


if __name__ == "__main__":
    args = create_parser(sys.argv[1:])
    checker = Checker("testing", args)
    sys.exit(checker.main())