#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
# Barcerlona Supercomputing Center - Centro Nacional de Supercomputacion
#        Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)
#             -
#             -
#

import logging, netCDF4, os
import dqc_wrapper.utils.grib_content as gcont
from dqc_wrapper.utils.commons import dir_or_file, initiate_logging

class SpatialCompletenessTest:
    def __init__(self, args):
        self.type = args.type
        self.path = args.input
        self.files = dir_or_file(self.path)


    def missing_values_netcdf(self, ncitem, var):
        filtered_var = ncitem.variables[var]
        variable = filtered_var[:]
        missing = variable.fill_value
        if not filtered_var.mask:
            # If the variable does not have a mask
            datamiss = len(variable[variable==missing])
        else:
            # If the variable does have a mask
            data_masked = variable.mask
            data = variable.data
            real_data = data[~data_masked]
            datamiss = len(real_data[real_data == missing])

        mk = datamiss > 0

        return datamiss, mk

    def netcdf_test(self, nc_list):
        output = list()

        for ncfile in nc_list:
            ncitem = netCDF4.Dataset(ncfile)
            ncitem.set_always_mask(True)
            test_var = [item for item in ncitem.variables.keys() \
                if item not in ncitem.dimensions.keys()]
            if test_var:
                for var in test_var:
                    datamiss, mk = self.missing_values_netcdf(ncitem, var)
                    if mk:
                        output_msg = ncfile.split('/')[-1] + ' variable '
                        output_msg += var + ' has missing data ' + datamiss
                        output.append('\n'.join([ncfile.split('/')[-1],
                                                 output_msg, var]))

        if not len(output):
            logging.info('\nNO GAPS FOUND! HURRAY!')

        else:
            logging.error('n' + '\n\n'.join(output))

    def missing_values_grib(self, msgs):
        # You should check how to see if a message has a mask
        output = list()
        for i, msg in enumerate(msgs):
            vals = msg['values']
            valmiss = msg['missingValue']
            numofmisscalc = len(vals[vals==valmiss])
            numofmiss = msg['numberOfMissing']

            if numofmisscalc == numofmiss and numofmiss > 0:
                output.append('Message number %s has %s num of missing values' %(i+1,numofmiss))
            
        if len(output):
            return '\n' + '\n'.join(output), 'ER'

        else:
            return '\nNo missing values', 'OK'

    def grib_test(self, grib_list):
        oute = list()
        for item in grib_list:
            gfile = gcont.grib_content(item)
            msgs = gfile.return_all_msg()

            string, wrn = self.missing_values_grib(msgs)
            print(wrn)
            if wrn == 'ER':
                stringA = '\n' + ' - '*20
                stringA += '\nFILE %s has missing values. More precisely\n\n' 
                stringA += string

            try:
                oute.append(stringA)
            except:
                stringA = oute

        if not len(oute):
            logging.info('\nNO GAPS FOUND! HURRAY!')

        else:
            logging.error('n' + '\n\n'.join(oute))


    def main(self):
        if self.type:
            initiate_logging(self.type,
                             'LOG_TEST02_SPATIAL_COMPLETENESS_CHECK',
                             self.path)
            if self.type == 'netcdf':
                ncfiles = [item for item in self.files if item.endswith('.nc')]
                if ncfiles:
                    self.netcdf_test(ncfiles)
                else:
                    raise Exception('\n\tNo NetCDF files in the given input\n')

            elif self.type == 'grib':
                gribfiles = [item for item in self.files if item.endswith('.grib')]
                if gribfiles:
                    self.grib_test(gribfiles)
                else:
                    raise Exception('\n\tNo GRIB files in the given input\n')
            else:
                raise Exception("Type argument not supported for this test")

        else:
            raise Exception('\n\tStandard type has not been given by the user\n')